#include "server.h"
#include "worker.h"
#include "listener.h"

CServer::CServer(const WORD nPort)
{
	m_wPort = nPort;
	m_pListener = NULL;
	m_hCompletion = NULL;
	m_dwThreadCount = 0;

	::InitializeCriticalSection(&m_csSocket);
	::InitializeCriticalSection(&m_csLoggedOn);
}

CServer::~CServer()
{
	// ͺ̽  
	FREE(m_pDb);

	// ʸ .
	FREE(m_pListener);

	//  Ŀ带 Ų.
	for (int n=0; n < (int)m_dwThreadCount; n++)
	{
		PostQueuedCompletionStatus(m_hCompletion, 0, NULL, NULL);
	}

	// Ŀ尴ü ޸𸮿 Ų.
	for (;;)
	{
		CWorker* pWorkerTemp = NULL;
		
		if (!m_deqWorker.empty())
		{
			pWorkerTemp = m_deqWorker.front();
			m_deqWorker.pop_front();
		}
		
		if (pWorkerTemp)
		{
			FREE(pWorkerTemp);
		}
		else
			break;
	}

	// ӵ  ڵ  ݰ Ҵ ޸𸮸 Ѵ.
	for (;;)
	{
		CUserContext* pUserContext = NULL;

		if (!m_deqSocket.empty())
		{
			pUserContext = m_deqSocket.front();
			m_deqSocket.pop_front();
		}
		else
			break;

		if (pUserContext)
		{
			::closesocket(pUserContext->Socket);
			FREE(pUserContext);
		}
		else
			break;
	}

	// øƮ ڵ ݴ´.
	::CloseHandle(m_hCompletion);
	// ũƼ  .
	::DeleteCriticalSection(&m_csSocket);
	::DeleteCriticalSection(&m_csLoggedOn);
}

BOOL CServer::Start()
{
	SYSTEM_INFO systemInfo;		// CPU  ˱
	
	// IOCP Ŀ  CPU*2 ŭ Ѵ.
	GetSystemInfo(&systemInfo);
	m_dwThreadCount = systemInfo.dwNumberOfProcessors*2; 	
	
	// ø Ʈ ڵ 
	m_hCompletion = CreateIoCompletionPort(INVALID_HANDLE_VALUE, NULL, 0, 0 );
	if (INVALID_HANDLE_VALUE == m_hCompletion )
	{
		CServer::WriteLog("IOCP Create Error");
		return FALSE;
	}

	// Ŀ  
	for (DWORD n=0; n<m_dwThreadCount; n++)
	{
		CWorker* pWorker = new CWorker(*this);
		m_deqWorker.push_back(pWorker);
	}

	m_pDb = new CDBManager();
	m_pDb->Connect(CONNECTION_PATH);

	//  
	m_pListener = new CListener(*this, m_wPort);

	// ʽ ..
	if (!m_pListener->Listen())
	{
		CServer::WriteLog("Listen Failed");
		return FALSE;
	}
	return TRUE;
}


void CServer::AcceptSocket(CUserContext* lpCompletionKey)
{
	::EnterCriticalSection(&m_csSocket);
	
	m_deqSocket.push_back(lpCompletionKey); 

	::LeaveCriticalSection(&m_csSocket);
}

// client   쿡 ȣȴ. 
void CServer::DeleteSocket(CUserContext* lpCompletionKey)
{
	::EnterCriticalSection(&m_csSocket);

	std::deque<CUserContext*>::iterator it = m_deqSocket.begin();

	for (; it != m_deqSocket.end(); it++)
	{
		CUserContext* pUserContext = *it;
		if (pUserContext == lpCompletionKey)
		{
			::closesocket(pUserContext->Socket);
			FREE(pUserContext);
			m_deqSocket.erase(it);
			break;
		}
	}

	::LeaveCriticalSection(&m_csSocket);
}

// α׿µ ڸƮ(map) , 
void CServer::AcceptLoggedOnUser(DWORD uniqueid, CUserContext* lpCompletionKey)
{
   	::EnterCriticalSection(&m_csLoggedOn);

	m_mapLoggedOn[uniqueid] = lpCompletionKey;

	::LeaveCriticalSection(&m_csLoggedOn);

}

void CServer::DeleteLoggedOnUser(DWORD uniqueid)
{
   	::EnterCriticalSection(&m_csLoggedOn);
	
	std::map<DWORD, CUserContext*>::iterator it;
	it = m_mapLoggedOn.find(uniqueid);	
	if ( it != m_mapLoggedOn.end() )
	{
		m_mapLoggedOn.erase(it);
	}

	::LeaveCriticalSection(&m_csLoggedOn);
}

// α׿µ  ˻
BOOL CServer::isLoggedOn(DWORD uniqueid)
{
	BOOL bRet;

	::EnterCriticalSection(&m_csLoggedOn);

	if ( m_mapLoggedOn.find(uniqueid) != m_mapLoggedOn.end() )
	{
		bRet = TRUE;
	}
	else
	{
		bRet = FALSE;
	}

	::LeaveCriticalSection(&m_csLoggedOn);

	return bRet;
}

// α׿µ  ؽƮ
CUserContext* CServer::GetLoggedOnUserContext(DWORD uniqueid)
{
	CUserContext* pUserContext;
    
	if (isLoggedOn(uniqueid) )
	{
		::EnterCriticalSection(&m_csLoggedOn);

		pUserContext = m_mapLoggedOn[uniqueid];

		::LeaveCriticalSection(&m_csLoggedOn);
	}
	else 
	{
		pUserContext = NULL;
	}

	return pUserContext;
}

size_t CServer::GetAcceptUser()
{
	size_t nRet;
	::EnterCriticalSection(&m_csSocket);

	nRet = m_deqSocket.size();

	::LeaveCriticalSection(&m_csSocket);

	return nRet;
}

size_t CServer::GetLogOnUser()
{
	size_t nRet;
	::EnterCriticalSection(&m_csLoggedOn);

	nRet = m_mapLoggedOn.size();

	::LeaveCriticalSection(&m_csLoggedOn);

	return nRet;
}

void CServer::WriteLog(const char* err, ...)
{
	char buf[1024];
	va_list ap;

	va_start(ap, err); //  .
	::vsprintf(buf, err, ap);

	// ó..
	std::cout << buf << std::endl;
	
	va_end(ap); //  볡..
}