#include "worker.h"
#include "server.h"

//////////////////////////////////////////////////////////////////////////
// Constructor / Destructor
//////////////////////////////////////////////////////////////////////////
CWorker::CWorker(CServer &rServer)
	: m_rServer(rServer)
{
	::InitializeCriticalSection(&m_csDb);
	m_bRun = TRUE;
	Start();
}

CWorker::~CWorker()
{
	::DeleteCriticalSection(&m_csDb);
	m_bRun = FALSE;
	WaitForTermination();
}


int CWorker::Run()
{
	CUserContext* pUserContext = NULL;
	PPER_IO_DATA lpPerIoData = NULL;
	DWORD dwBytesTransferred = 0;

	while (m_bRun)
	{
		BOOL bRet = ::GetQueuedCompletionStatus( m_rServer.m_hCompletion,
												&dwBytesTransferred,
												(LPDWORD)&pUserContext,
												(LPOVERLAPPED*)&lpPerIoData,
												INFINITE);
		if (bRet)
		{
			if (NULL == pUserContext) // 
			{
				break;
			}
		}
		else
		{
			// ̰ ڵ 64 κ̴.
			// ڵ 64  Ʈũ̸ Ҽ ° ̹Ƿ  ݴ´.
			CServer::WriteLog("Disconnect Socket[%d]", pUserContext->Socket);
			m_rServer.DeleteSocket(pUserContext);
			continue;
		}

		if (NULL == lpPerIoData )
		{
			CServer::WriteLog("Getting Completion Packet Fail : %d Error ", ::GetLastError());
			continue;
		}

		try
		{
			// Ŭ̾Ʈ   
			if(0 == dwBytesTransferred)
			{
				throw "dwBytestransferred==0";
			}

			// IO ݿ  ׿  ó
			if(lpPerIoData == pUserContext->pRecvData)
			{
				// RECV Operation 
			    if(!RecvCompleteEvent(pUserContext, dwBytesTransferred))	
				{
					throw "RecvCompleteEvent Error";					
				}				
			}
			else if(lpPerIoData == pUserContext->pSendData)
			{
				// SEND Operation
                if(!SendCompleteEvent(pUserContext, dwBytesTransferred))	
				{
					throw "SendCompleteEvent Error";				
				}			
			}
			else
			{				
				if(!OtherCompleteEvent(pUserContext, dwBytesTransferred))	
				{
					throw "OtherCompleteEvent Error";					
				}	
			}
			
		}
		catch(char* errText)
		{			
			CServer::WriteLog("Dissconnect Socket[%d] : %s", pUserContext->Socket, errText);
				// continue   Abnormal Termination  ȴ.
				m_rServer.DeleteSocket(pUserContext);
		}
	}  // while 

	return 0;
}


// Recv û
BOOL CWorker::RecvPost(CUserContext* pUserContext)
{
	DWORD dwRecvBytes=0;
	DWORD dwFlags=0;
	
	pUserContext->InitRecvData();
    
	int ret = WSARecv(pUserContext->Socket,&(pUserContext->pRecvData->wsabuf),1,
		    &dwRecvBytes,&dwFlags,&(pUserContext->pRecvData->overlapped),NULL);

	if (SOCKET_ERROR == ret)
	{
		if (WSA_IO_PENDING != ::WSAGetLastError())
		{			
			CServer::WriteLog("Recv Request Error(WSARecv Function Failed): %d", ::WSAGetLastError());;
			return FALSE;
		}
	}

	return TRUE;
}

// Send û
BOOL CWorker::SendPost(CUserContext* pUserContext)
{
	DWORD dwFlags=0;
	DWORD dwSendBytes=0;
	
	::memset(&pUserContext->pSendData->overlapped, 0, sizeof(OVERLAPPED));

	int ret = ::WSASend(pUserContext->Socket, 
					&(pUserContext->pSendData->wsabuf), 1,
					&dwSendBytes, dwFlags,
					&(pUserContext->pSendData->overlapped),NULL);

	if (SOCKET_ERROR == ret)
	{
		if (WSA_IO_PENDING != ::WSAGetLastError())
		{			
			CServer::WriteLog("Send Request Error(WSASend Function Failed): %d", ::WSAGetLastError());;
			return FALSE;
		}
	}

	return TRUE;
}

// ú ̺Ʈ ó ڵ鷯 Լ
BOOL CWorker::RecvCompleteEvent(CUserContext* pUserContext, DWORD dwBytesTransferred)
{
	BOOL bLogOn;
//	BOOL bMultiPacket = FALSE;	// #MULTI ߸ Ŷ ɼ Ͽ ӽ
	CPacket* pPacket = &(pUserContext->pRecvData->packet);

	// α׿   ƴ 
	if (0 == pUserContext->nUniqueid)
	{
		bLogOn = FALSE;
	}
	else
	{
		CUserContext* pCheckUser = m_rServer.GetLoggedOnUserContext(pUserContext->nUniqueid);
		if (NULL == pCheckUser)
		{
			bLogOn = FALSE;
		}
		else
		{
			if (pCheckUser->Socket == pUserContext->Socket)
				bLogOn = TRUE;
			else 
				bLogOn = FALSE;
		}
	}

	// Ŷȣ ùٸ 
	if (pPacket->id != (pUserContext->pRecvData->nNum + 1) )
	{
		/*** ޽  ʰ ٷ .

		// ùٸ    ޽  Ϸ̺Ʈ  ݴ´.
		pUserContext->bCloseSocket = TRUE;
		
		pUserContext->UpdateSendData(DISCONNECTION, INVALID_PACKET_ID);

		SendPost(pUserContext);
		return TRUE;
		**/

		CServer::WriteLog("Socket[%d] : Invalid Packet ID", pUserContext->Socket);
		return FALSE;
	}
	else
	{
		pUserContext->pRecvData->nNum = pPacket->id;
	}

	/**		#MULTI ߸ Ŷ ɼ Ͽ ӽ

	// Ŷ  Ŷ ƴ  Ŷ ɰϰ..
	if (pPacket->type != 0)
	{
		//  Ŷ 
		if (pPacket->packetnum == 1)
		{
			pUserContext->pRecvData->pSumData = new char[pPacket->type * PACKET_DATA_SIZE];
			pUserContext->pRecvData->PacketNum = pPacket->type;
			::memcpy(pUserContext->pRecvData->pSumData, pPacket->data, PACKET_DATA_SIZE);
			pUserContext->pRecvData->SumDataLen = PACKET_DATA_SIZE;

			return TRUE;
		}
		//  Ŷϰ
		else if (pPacket->packetnum == pUserContext->PacketNum)
		{
			::memcpy(pUserContext->pRecvData->pSumData + ( (pPacket->packetnum-1) * PACKET_DATA_SIZE),
					pPacket->data, pPacket->length);
			pUserContext->pRecvData->SumDataLen += pPacket->length;

			bMultiPacket = TRUE;
		}
		// ߰ Ŷϰ
		else if (pPacket->packetnum < pUserContext->PacketNum)
		{
			::memcpy(pUserContext->pRecvData->pSumData + ( (pPacket->packetnum-1) * PACKET_DATA_SIZE),
					pPacket->data, PACKET_DATA_SIZE);
			pUserContext->pRecvData->SumDataLen += PACKET_DATA_SIZE;

			return TRUE;
		}
		// ߸ Ŷȣ
		else
		{
			CServer::WriteLog("Socket[%d] : Invalid Multi Packet Number", pUserContext->Socket);
			return FALSE;
		}
	}

	**/

	//  ó...
	switch(pPacket->command)
	{
		//  üũ
	case VER:
		//  Ŷ ϰ
		if (0 != pPacket->type)
		{
			CServer::WriteLog("Socket[%d] Command:VER, Recv MultiPacket");
			return FALSE;
		}
		
		if (0 == ::strncmp(VERSION, pPacket->data, pPacket->length) )
		{
			pUserContext->UpdateSendData(VER, OK);
			SendPost(pUserContext);
			RecvPost(pUserContext);
			return TRUE;
		}
		else
		{	
			//    α   ,  
			CServer::WriteLog("Socket[%d] Invalid Version", pUserContext->Socket);
			pUserContext->CloseSocket = TRUE;
			pUserContext->UpdateSendData(VER, INVALID_VER);
			SendPost(pUserContext);
			return TRUE;
		}

		break;

	case LOGON:
		if (0 != pPacket->type)
		{
			CServer::WriteLog("Socket[%d] Command:LOGON, Recv MultiPacket");
			return FALSE;
		}
		char userID[30];
		char passwd[30];

		//  ̵, н带 иѴ.

		break;
	case LOGOUT:
		break;
	case USERINFO:
		break;
	case SIGNUP:
		break;
	case UNSIGN:
		break;
	default:
		break;
	}

	return TRUE;
}

// Send Ϸ Ŷ ó ڵ鷯 Լ
BOOL CWorker::SendCompleteEvent(CUserContext* pUserContext, DWORD dwBytesTransferred)
{
	pUserContext->pSendData->nCurBytes += dwBytesTransferred;

	// Ͱ  ۵ ʾ
	if (pUserContext->pSendData->nCurBytes < pUserContext->pSendData->nTotalBytes )
	{
		pUserContext->pSendData->wsabuf.buf = pUserContext->pSendData->packet.szBuffer + 
												pUserContext->pSendData->nCurBytes;
		pUserContext->pSendData->wsabuf.len = pUserContext->pSendData->nTotalBytes -
												pUserContext->pSendData->nCurBytes;
		BOOL bRetval = SendPost(pUserContext);
		if (FALSE == bRetval)
			return FALSE;
	}
	else	// Ͱ  ۵Ǿ .
	{
		// FALSE ϸ  ݴ´.
		if (pUserContext->bCloseSocket)
		{
			return FALSE;
		}

		pUserContext->InitSendData();
	}

	return TRUE;
}

// Recv, Send Ϸ   ó ڵ鷯 Լ
BOOL CWorker::OtherCompleteEvent(CUserContext* pUserContext, DWORD dwBytesTransferred)
{
	return FALSE;
}